from setuptools import setup

setup(name='rManifolds',
      version='0.1',
      description='Manifolds in snappy from random knots/links/braids',
      url='http://github.com/mauriciogtec/rManifolds',
      author='Mauricio B. Garcia Tec',
      license='MIT',
      packages=['rManifolds'],
      zip_safe=False)
